import random
from pathlib import Path

import numpy as np
import scipy.ndimage as ndi
import SimpleITK as sitk
import torch

# === Processing Settings ===
SIZE_MM = 70
SIZE_PX = 518

# === Training Parameters ===
SEED = 42
NUM_WORKERS = 8

# === Data Augmentation ===
ROTATION = ((-20, 20), (-20, 20), (-20, 20))
TRANSLATION = True

def set_seed(seed=42):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

def apply_augmentation(slice_2d):
    if ROTATION:
        angles = [np.random.uniform(*ROTATION[i]) for i in range(3)]
        slice_2d = ndi.rotate(slice_2d, angle=angles[0], axes=(0, 1), reshape=False, mode='nearest')
    if TRANSLATION:
        shift_values = np.random.uniform(-5, 5, size=2)
        slice_2d = ndi.shift(slice_2d, shift=shift_values, mode='nearest')
    return slice_2d

def clip_and_scale(npzarray, maxHU=400.0, minHU=-1000.0):
    npzarray = (npzarray - minHU) / (maxHU - minHU)
    npzarray = np.clip(npzarray, 0, 1)
    return npzarray

class CTScanDataset(torch.utils.data.Dataset):
    def __init__(self, scan_path: Path, seed: int = 42):
        self.SIZE_MM = SIZE_MM
        self.SIZE_PX = SIZE_PX
        self.SEED = SEED
        self.NUM_WORKERS = NUM_WORKERS

        set_seed(seed)
        self.scan_path = scan_path
        self.case_id = scan_path.stem

        image_sitk = sitk.ReadImage(str(scan_path))
        self.img = clip_and_scale(sitk.GetArrayFromImage(image_sitk))
        self.img = np.array([apply_augmentation(slice_2d) for slice_2d in self.img])

    def __getitem__(self, idx):
        return {
            "image": torch.from_numpy(self.img).float(),
            "ID": self.case_id,
            "image_path": str(self.scan_path)
        }

    def __len__(self):
        return 1

def get_scan_dataset(scan_path: Path, seed: int = 42):
    return CTScanDataset(scan_path, seed)



'''
    Adapted from https://github.com/mlfoundations/open_clip/blob/main/src/open_clip/transformer.py
                 https://github.com/bytedance/ibot/blob/main/models/vision_transformer.py 
                 https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/vision_transformer.py
'''
import math
from functools import partial
from typing import Callable

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint
from timm.layers import DropPath, Mlp, trunc_normal_
from timm.models._manipulate import checkpoint_seq, named_apply
from timm.models.vision_transformer import get_init_weights_vit
from einops import repeat

from .configuration_titan import TitanVisionConfig

def preprocess_features(features: torch.Tensor, coords: torch.Tensor, patch_size_lv0: int):
    # Remove extra dimensions
    features = features.squeeze(0) if features.dim() == 3 else features
    coords = coords.squeeze(0) if coords.dim() == 3 else coords

    # Offset and normalize coordinates
    offset = coords.min(dim=0).values
    grid_coords = torch.floor_divide(coords - offset, patch_size_lv0)

    # Compute grid size
    grid_offset = grid_coords.min(dim=0).values
    grid_coords = grid_coords - grid_offset
    _H, _W = grid_coords.max(dim=0).values + 1

    # Create feature and coordinate grids
    feature_grid = torch.zeros((_H, _W, features.size(-1)), device=features.device)
    coords_grid = torch.zeros((_H, _W, 2), dtype=torch.int64, device=coords.device)

    # Use scatter for more efficient placement
    indices = grid_coords[:, 0] * _W + grid_coords[:, 1]
    feature_grid.view(-1, features.size(-1)).index_add_(0, indices, features)
    coords_grid.view(-1, 2).index_add_(0, indices, coords)

    # Permute grids
    feature_grid = feature_grid.permute(2, 0, 1)
    coords_grid = coords_grid.permute(2, 0, 1)

    # Background mask
    bg_mask = torch.any(feature_grid != 0, dim=0)
    return feature_grid.unsqueeze(0), coords_grid.unsqueeze(0), bg_mask.unsqueeze(0)

class Attention(nn.Module):
    def __init__(
            self,
            dim,
            num_heads=8,
            qkv_bias=False,
            qk_norm=False,
            attn_drop=0.,
            proj_drop=0.,
            norm_layer=nn.LayerNorm,
            pos_encode: str = None,
    ):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim ** -0.5
        self.pos_encode = pos_encode

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.q_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.k_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.attn_drop_prob = attn_drop

    def forward(self, x, attn_bias, bg_mask=None):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)
        q, k = self.q_norm(q), self.k_norm(k)

        if self.pos_encode == 'alibi':
            if bg_mask is not None and B > 1:
                bg_mask = bg_mask.view(B, -1)
                bg_mask = torch.cat((torch.ones((B, 1), dtype=bg_mask.dtype, device=bg_mask.device), bg_mask), dim=-1)
                attn_mask = bg_mask.unsqueeze(2) * bg_mask.unsqueeze(1)
                # add entries on the diagonal to avoid entire rows being False                
                diag_mask = torch.eye(attn_mask.size(1), device=attn_mask.device, dtype=torch.bool).unsqueeze(0)
                attn_mask = torch.logical_or(attn_mask, diag_mask)
                attn_mask = (1 - attn_mask.float()) * torch.finfo(q.dtype).min
                attn_mask = repeat(attn_mask, 'b i j -> b h i j', h=self.num_heads)
                attn_mask = attn_mask + attn_bias
                x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask, dropout_p=self.attn_drop_prob)
            else:
                x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_bias, dropout_p=self.attn_drop_prob)
        elif self.pos_encode in ['none', 'abs']:
            if bg_mask is not None and B > 1:
                bg_mask = bg_mask.view(B, -1)
                bg_mask = torch.cat((torch.ones((B, 1), dtype=bg_mask.dtype, device=bg_mask.device), bg_mask), dim=-1)
                # add entries on the diagonal to avoid entire rows being False                
                attn_mask = bg_mask.unsqueeze(2) * bg_mask.unsqueeze(1)
                diag_mask = torch.eye(attn_mask.size(1), device=attn_mask.device, dtype=torch.bool).unsqueeze(0)
                attn_mask = torch.logical_or(attn_mask, diag_mask)
                attn_mask = (1 - attn_mask.float()) * torch.finfo(q.dtype).min
                x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask.unsqueeze(1), dropout_p=self.attn_drop_prob)
            else:
                x = F.scaled_dot_product_attention(q, k, v, dropout_p=self.attn_drop_prob)
        else:
            raise ValueError(f'pos_encode {self.pos_encode} not supported')

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class LayerScale(nn.Module):
    def __init__(self, dim, init_values=1e-5, inplace=False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x):
        return x.mul_(self.gamma) if self.inplace else x * self.gamma


class Block(nn.Module):
    def __init__(
            self,
            dim,
            num_heads,
            mlp_ratio=4.,
            qkv_bias=False,
            qk_norm=False,
            proj_drop=0.,
            attn_drop=0.,
            init_values=None,
            drop_path=0.,
            act_layer=nn.GELU,
            norm_layer=nn.LayerNorm,
            mlp_layer=Mlp,
            pos_encode=None,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_norm=qk_norm,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
            norm_layer=norm_layer,
            pos_encode=pos_encode,
        )
        self.ls1 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        self.drop_path1 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.norm2 = norm_layer(dim)
        self.mlp = mlp_layer(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            act_layer=act_layer,
            drop=proj_drop,
        )
        self.ls2 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        self.drop_path2 = DropPath(drop_path) if drop_path > 0. else nn.Identity()
    
    def forward(self, x, attn_bias, bg_mask=None):
        x = x + self.drop_path1(self.ls1(self.attn(self.norm1(x), attn_bias=attn_bias, bg_mask=bg_mask)))
        x = x + self.drop_path2(self.ls2(self.mlp(self.norm2(x))))
        return x

class CustomSequential(nn.Module):
    def __init__(self, *modules):
        super(CustomSequential, self).__init__()
        self.modules_list = nn.ModuleList(modules)

    def forward(self, x, attn_mask, bg_mask=None):
        for i, module in enumerate(self.modules_list):
            x = module(x, attn_mask, bg_mask)
        return x

class AttentionalPooler(nn.Module):
    def __init__(
            self,
            d_model: int,
            context_dim: int,
            n_head: int = 8,
            n_queries: int = 256,
            norm_layer: Callable = nn.LayerNorm
    ):
        super().__init__()
        self.query = nn.Parameter(torch.randn(n_queries, d_model))
        self.attn = nn.MultiheadAttention(d_model, n_head, kdim=context_dim, vdim=context_dim)
        self.ln_q = norm_layer(d_model)
        self.ln_k = norm_layer(context_dim)
        self.n_head = n_head

    def forward(self, x: torch.Tensor, bg_mask: torch.Tensor = None):
        B = x.shape[0]
        if bg_mask is not None and B > 1:
            bg_mask = bg_mask.view(B, -1)
            bg_mask = torch.cat((torch.ones((B, 1), dtype=bg_mask.dtype, device=bg_mask.device), bg_mask), dim=-1)
            bg_mask = bg_mask.unsqueeze(1).repeat(self.n_head, self.query.shape[0], 1)
            bg_mask = ~bg_mask
        x = self.ln_k(x).permute(1, 0, 2).contiguous()  # NLD -> LND
        N = x.shape[1]
        q = self.ln_q(self.query)
        if B == 1:
            out, _ = self.attn(q.unsqueeze(1).expand(-1, N, -1), x, x, need_weights=False)
        else:
            out = self.attn(q.unsqueeze(1).expand(-1, N, -1), x, x, attn_mask=bg_mask, need_weights=False)[0]
        return out.permute(1, 0, 2).contiguous()  # LND -> NLD

class VisionTransformer(nn.Module):
    """ Vision Transformer
    A PyTorch impl of : `An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale`
        - https://arxiv.org/abs/2010.11929
    """

    def __init__(
            self,
            grid_size=14,
            num_classes=0,
            global_pool='token',
            embed_dim=768,
            depth=12,
            num_heads=12,
            mlp_ratio=4.,
            qkv_bias=True,
            qk_norm=False,
            init_values=None,
            class_token=True,
            no_embed_class=False,
            pre_norm=False,
            fc_norm=None,
            drop_rate=0.,
            attn_drop_rate=0.,
            drop_path_rate=0.,
            weight_init='',
            norm_layer=None,
            act_layer=None,
            block_fn=Block,
            return_all_tokens=False, 
            masked_im_modeling=False,
            mlp_patch_embed_dim=768,
            pos_encode_type='alibi',
            #### CoCa params ####
            attentional_pool: str = None,
            attn_pooler_queries: int = 128,
            attn_pooler_heads: int = 8
    ):
        """
        Args:
            grid_size (int): input feature grid size
            num_classes (int): number of classes for classification head
            global_pool (str): type of global pooling for final sequence (default: 'token')
            embed_dim (int): embedding dimension
            depth (int): depth of transformer
            num_heads (int): number of attention heads
            mlp_ratio (int): ratio of mlp hidden dim to embedding dim
            qkv_bias (bool): enable bias for qkv if True
            init_values: (float): layer-scale init values
            class_token (bool): use class token
            fc_norm (Optional[bool]): pre-fc norm after pool, set if global_pool == 'avg' if None (default: None)
            drop_rate (float): dropout rate
            attn_drop_rate (float): attention dropout rate
            drop_path_rate (float): stochastic depth rate
            weight_init (str): weight init scheme
            norm_layer: (nn.Module): normalization layer
            act_layer: (nn.Module): MLP activation layer
        """
        super().__init__()
        assert global_pool in ('', 'avg', 'token')
        assert class_token or global_pool != 'token'
        use_fc_norm = global_pool == 'avg' if fc_norm is None else fc_norm
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        act_layer = act_layer or nn.GELU
        self.return_all_tokens = return_all_tokens     # from ibot
        self.masked_im_modeling = masked_im_modeling   # from ibot

        self.num_classes = num_classes
        self.global_pool = global_pool
        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models
        self.num_prefix_tokens = 1 if class_token else 0
        self.no_embed_class = no_embed_class
        self.grad_checkpointing = False

        ## change embed_layer to mlp
        self.patch_embed = nn.Sequential(
            nn.Linear(mlp_patch_embed_dim, embed_dim),
            nn.GELU()
        )

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim)) if class_token else None
        # instead of embed_len = num_patches, we use grid_size
        self.pos_encode_type = pos_encode_type
        self.num_heads = num_heads
        self.pos_embed = nn.Parameter(torch.randn(1, grid_size ** 2 + self.num_prefix_tokens, embed_dim)) if pos_encode_type not in ['none', 'alibi'] else None
        self.pos_drop = nn.Dropout(p=drop_rate)
        if pos_encode_type == 'alibi':
            self.local_alibi = self.get_alibi(6, 6)
            self.global_alibi = self.get_alibi(14, 14)
            self.global_alibi_status = False
            self.local_alibi_status = False

        self.norm_pre = norm_layer(embed_dim) if pre_norm else nn.Identity()

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = CustomSequential(*[
            block_fn(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_norm=qk_norm,
                init_values=init_values,
                proj_drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[i],
                norm_layer=norm_layer,
                act_layer=act_layer,
                pos_encode=pos_encode_type,
            )
            for i in range(depth)])
        
        self.norm = norm_layer(embed_dim) if not use_fc_norm else nn.Identity()

        # classifier head
        self.fc_norm = norm_layer(embed_dim) if use_fc_norm else nn.Identity()
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

        if weight_init != 'skip':
            self.init_weights(weight_init)
        
        self.masked_im_modeling = masked_im_modeling
        if masked_im_modeling:
            self.masked_embed = nn.Parameter(torch.zeros(1, embed_dim))
        
        if attentional_pool:
            self.attn_pool_type = attentional_pool
            self.pool_type = 'none'
            if attentional_pool in ('parallel', 'cascade'):
                self.attn_pool = AttentionalPooler(
                    embed_dim,
                    embed_dim,
                    n_head=attn_pooler_heads,
                    n_queries=attn_pooler_queries,
                )
                self.attn_pool_contrastive = AttentionalPooler(
                    embed_dim,
                    embed_dim,
                    n_head=attn_pooler_heads,
                    n_queries=1,
                )
            else:
                raise NotImplementedError
            pool_dim = embed_dim
            scale = embed_dim ** -0.5
            self.proj = nn.Parameter(scale * torch.randn(pool_dim, pool_dim))
        else:
            self.attn_pool = None
            self.attn_pool_contrastive = None

    def init_weights(self, mode=''):
        assert mode in ('jax', 'jax_nlhb', 'moco', '')
        head_bias = -math.log(self.num_classes) if 'nlhb' in mode else 0.
        if self.pos_encode_type not in ['none', 'alibi']:
            trunc_normal_(self.pos_embed, std=.02)
        if self.cls_token is not None:
            nn.init.normal_(self.cls_token, std=1e-6)
        named_apply(get_init_weights_vit(mode, head_bias), self)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed', 'cls_token', 'dist_token'}

    @torch.jit.ignore
    def group_matcher(self, coarse=False):
        return dict(
            stem=r'^cls_token|pos_embed|patch_embed',  # stem and embed
            blocks=[(r'^blocks\.(\d+)', None), (r'^norm', (99999,))]
        )

    @torch.jit.ignore
    def set_grad_checkpointing(self, enable=True):
        self.grad_checkpointing = enable

    @torch.jit.ignore
    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes: int, global_pool=None):
        self.num_classes = num_classes
        if global_pool is not None:
            assert global_pool in ('', 'avg', 'token')
            self.global_pool = global_pool
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

    def _pos_embed(self, x, coords, w, h):
        if self.no_embed_class:
            # deit-3, updated JAX (big vision)
            # position embedding does not overlap with class token, add then concat
            x = x + self.interpolate_pos_encoding(x, coords, w, h)
            if self.cls_token is not None:
                x = torch.cat((self.cls_token.expand(x.shape[0], -1, -1), x), dim=1)
        else:
            # original timm, JAX, and deit vit impl
            # pos_embed has entry for class token, concat then add
            if self.cls_token is not None:
                x = torch.cat((self.cls_token.expand(x.shape[0], -1, -1), x), dim=1)
            x = x + self.interpolate_pos_encoding(x, coords, w, h)
        return self.pos_drop(x)

    def forward_features(self, x, coords=None, mask=None, bg_mask=None):
        B, nc, w, h = x.shape
        ## flatten x before pass into mlp patch embed
        x = x.flatten(2, 3).transpose(1,2)
        if self.pos_encode_type == 'alibi':
            if w*h == 36 and B != 1:
                if not self.local_alibi_status:
                    self.prepare_tensor(x, 'local', 'alibi')
                attn_bias = self.local_alibi
            elif w*h == 196 and B != 1:
                if not self.global_alibi_status:
                    self.prepare_tensor(x, 'global', 'alibi')
                attn_bias = self.global_alibi
            else:
                attn_bias = self.get_alibi(w, h, bg_mask) if B == 1 else self.get_alibi(w, h)
                attn_bias = attn_bias.repeat(x.shape[0], 1, 1, 1).type(x.dtype).to(x.device)
        else:
            attn_bias = None
            
        if self.masked_im_modeling:
            assert mask is not None
            x = self.patch_embed(x)
            x = self.mask_model(x, mask)
        else:
            x = self.patch_embed(x)
        
        x = self._pos_embed(x, coords, w, h)
        x = self.norm_pre(x)
        
        # mask background tokens when evaluating (batch size = 1)
        if bg_mask is not None and B == 1:
            bg_mask = torch.cat((torch.ones((1, 1), dtype=torch.bool, device=x.device), bg_mask.view(1, -1)), dim=1)
            x = x[bg_mask].unsqueeze(0)

        if self.grad_checkpointing and not torch.jit.is_scripting():
            x = checkpoint_seq(self.blocks, x, attn_bias, bg_mask)
        else:
            x = self.blocks(x, attn_bias, bg_mask)
        
        x = self.norm(x)
        return x

    def forward_head(self, x, pre_logits: bool = False):
        if self.global_pool:
            x = x[:, self.num_prefix_tokens:].mean(dim=1) if self.global_pool == 'avg' else x[:, 0]
        x = self.fc_norm(x)
        return x if pre_logits else self.head(x)

    def forward_attn_pool(self, x, bg_mask=None):
        tokens = self.attn_pool(x, bg_mask=bg_mask)
        if self.attn_pool_type == 'parallel':
            pooled = self.attn_pool_contrastive(x, bg_mask=bg_mask)
        else:
            assert self.attn_pool_type == 'cascade'
            pooled = self.attn_pool_contrastive(tokens, bg_mask=bg_mask)
        pooled = pooled.squeeze(1)
        return pooled, tokens

    def forward(self, patch_features, patch_coords, patch_size_lv0, return_all_tokens=False, mask=None, bg_mask=None, no_proj=False):
        x, coords, bg_mask = preprocess_features(patch_features, patch_coords, patch_size_lv0)
        if self.masked_im_modeling:
            assert mask is not None
            x = self.forward_features(x, coords=coords, mask=mask, bg_mask=bg_mask)
        else:
            x = self.forward_features(x, coords=coords, bg_mask=bg_mask)
        
        if self.attn_pool is not None: ### CoCa
            pooled, tokens = self.forward_attn_pool(x, bg_mask=bg_mask)
            if no_proj:
                return pooled
            pooled = pooled @ self.proj
            if self.return_all_tokens:
                return pooled, tokens
            else:
                return pooled
        else: ### iBOT
            return_all_tokens = self.return_all_tokens if \
                return_all_tokens is None else return_all_tokens
            if return_all_tokens:
                return x
            x = self.forward_head(x)
            return x
    
    def interpolate_pos_encoding(self, x, coords, w, h):
        if self.pos_encode_type in ['none', 'alibi']:
            return torch.zeros_like(x)
        elif self.pos_encode_type == 'abs':
            # print('Using 2d absolute pos encode')
            npatch = x.shape[1] - 1
            N = self.pos_embed.shape[1] - 1
            if npatch == N and w == h:
                return self.pos_embed
            class_pos_embed = self.pos_embed[:, 0]
            patch_pos_embed = self.pos_embed[:, 1:]
            dim = x.shape[-1]
            ## MODIFICATIONS: using w // 1 and h // 1 instead of w // patch_size 
            w0 = w // 1
            h0 = h // 1
            # we add a small number to avoid floating point error in the interpolation
            # see discussion at https://github.com/facebookresearch/dino/issues/8
            w0, h0 = w0 + 0.1, h0 + 0.1
            patch_pos_embed = nn.functional.interpolate(
                patch_pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
                scale_factor=(w0 / math.sqrt(N), h0 / math.sqrt(N)),
                mode='bicubic',
            )
            assert int(w0) == patch_pos_embed.shape[-2] and int(h0) == patch_pos_embed.shape[-1]
            patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
            return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1)
        else:
            raise NotImplementedError

    def mask_model(self, x, mask):
        mask = mask.view(mask.shape[0], -1)
        x[mask, :] = self.masked_embed.to(x.dtype)
        return x

    def get_alibi(self, w, h, bg_mask=None):
        x, y = np.meshgrid(np.arange(w), np.arange(h), indexing='ij')
        if bg_mask is not None:
            x = x[bg_mask.cpu().squeeze(0)]
            y = y[bg_mask.cpu().squeeze(0)]
        points = np.stack([x.ravel(), y.ravel()], axis=1)  # (w*h, 2)
        diffs = points[:, None, :] - points[None, :, :]
        dists = np.sqrt(np.sum(diffs ** 2, axis=-1))  # (w*h, w*h)
        def get_slopes(n):
            if math.log2(n).is_integer():
                p = 2 ** (-2 ** -(math.log2(n) - 3))
                return [p * (p ** i) for i in range(n)]
            nearest_power_of_2 = 2 ** math.floor(math.log2(n))
            base_slopes = get_slopes(nearest_power_of_2)
            if nearest_power_of_2 == n:
                return base_slopes
            extra_slopes = get_slopes(2 * nearest_power_of_2)[0::2][:n - nearest_power_of_2]
            return base_slopes + extra_slopes
        
        slopes = torch.tensor(get_slopes(self.num_heads), dtype=torch.float32).view(self.num_heads, 1, 1)
        n_patches = dists.shape[-1]  # w*h or bg_mask.sum()
        dists_tensor = torch.tensor(dists, dtype=torch.float32).view(1, n_patches, n_patches)
        bias_matrix = dists_tensor * slopes * -1  # (1, num_heads, w*h, w*h)
        embed_len = n_patches + 1
        all_bias = torch.zeros(1, self.num_heads, embed_len, embed_len)
        all_bias[:, :, 1:, 1:] = bias_matrix
        return all_bias
    
    def prepare_tensor(self, x, view, type):
        if view == 'local':
            if type == 'alibi':
                self.local_alibi = self.local_alibi.repeat(x.shape[0], 1, 1, 1).type(x.dtype).to(x.device)
                self.local_alibi_status = True
        elif view == 'global':
            if type == 'alibi':
                self.global_alibi = self.global_alibi.repeat(x.shape[0], 1, 1, 1).type(x.dtype).to(x.device)
                self.global_alibi_status = True
        else: 
            raise NotImplementedError

def build_vision_tower(
        vision_cfg: TitanVisionConfig
    ):
    vision = VisionTransformer(
        grid_size=vision_cfg.grid_size,
        global_pool=vision_cfg.global_pool,
        embed_dim=vision_cfg.embed_dim,
        depth=vision_cfg.depth,
        num_heads=vision_cfg.num_heads,
        mlp_ratio=vision_cfg.mlp_ratio,
        qkv_bias=vision_cfg.qkv_bias,
        mlp_patch_embed_dim=vision_cfg.mlp_patch_embed_dim,
        pos_encode_type=vision_cfg.pos_encode_type,
        attentional_pool=vision_cfg.attentional_pool,
        attn_pooler_queries=vision_cfg.attn_pooler_queries,
        attn_pooler_heads=vision_cfg.attn_pooler_heads
    )
    return vision


import copy
from typing import Any

from transformers import PretrainedConfig

class ConchConfig(PretrainedConfig):
    model_type = "conch"

    def __init__(
        self,
        patch_size: int = 16,
        context_dim: int = 1024,
        embed_dim: int = 768,
        depth: int = 24,
        num_heads: int = 16,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        init_values: float = 1e-6,
        pooler_n_queries_contrast: int = 1,
        **kwargs: Any,
    ):
        self.patch_size = patch_size
        self.context_dim = context_dim
        self.embed_dim = embed_dim
        self.depth = depth
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.init_values = init_values
        self.pooler_n_queries_contrast = pooler_n_queries_contrast

        super().__init__(**kwargs)

class TitanVisionConfig(PretrainedConfig):
    model_type = "titan_vision"

    def __init__(
        self,
        grid_size: int =14,
        global_pool: str ='token',
        embed_dim: int = 768,
        depth: int = 6,
        num_heads: int = 12,
        mlp_ratio: float = 4.,
        qkv_bias: bool = True,
        mlp_patch_embed_dim: int =768,
        pos_encode_type: str ='alibi',
        #### CoCa params ####
        attentional_pool: str = None,
        attn_pooler_queries: int = 128,
        attn_pooler_heads: int = 8,
        **kwargs: Any,
    ):
        self.grid_size = grid_size
        self.global_pool = global_pool
        self.embed_dim = embed_dim
        self.depth = depth
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.mlp_patch_embed_dim = mlp_patch_embed_dim
        self.pos_encode_type = pos_encode_type
        self.attentional_pool = attentional_pool
        self.attn_pooler_queries = attn_pooler_queries
        self.attn_pooler_heads = attn_pooler_heads

        super().__init__(**kwargs)

class TitanTextConfig(PretrainedConfig):
    model_type = "titan_text"

    def __init__(
        self,
        embed_dim: int = 768,
        context_length: int = 128,
        vocab_size: int = 32007,

        width: int = 768,
        heads: int = 12,
        layers: int = 12,
        mlp_ratio: float = 4.0,
        embed_cls: bool = True,
        pad_id: int = 0,
        pool_type: str = 'argmax',
        proj_bias: bool = False,
        output_tokens: bool = True,
        **kwargs: Any,
    ):
        self.embed_dim = embed_dim
        self.context_length = context_length
        self.vocab_size = vocab_size

        self.width = width
        self.heads = heads
        self.layers = layers
        self.mlp_ratio = mlp_ratio
        self.embed_cls = embed_cls
        self.pad_id = pad_id
        self.pool_type = pool_type
        self.proj_bias = proj_bias
        self.output_tokens = output_tokens

        super().__init__(**kwargs)

class TitanConfig(PretrainedConfig):
    model_type = "titan"
    is_composition = True

    def __init__(
        self,
        vision_config: TitanVisionConfig = TitanVisionConfig(),
        text_config: TitanTextConfig = TitanTextConfig(),
        conch_config: ConchConfig = ConchConfig(),
        **kwargs: Any,
    ):
        if isinstance(conch_config, dict):
            self.conch_config = ConchConfig(**conch_config)
        else:
            self.conch_config = conch_config
        if isinstance(vision_config, dict):
            self.vision_config = TitanVisionConfig(**vision_config)
        else:
            self.vision_config = vision_config
        ### for CoCa ###
        self.vision_config.attentional_pool = "parallel"
        if isinstance(text_config, dict):
            self.text_config = TitanTextConfig(**text_config)
        else:
            self.text_config = text_config

        super().__init__(**kwargs)

    def to_dict(self):
        output = copy.deepcopy(self.__dict__)
        for k, v in output.items():
            if isinstance(v, PretrainedConfig):
                output[k] = v.to_dict()
        output["model_type"] = self.__class__.model_type
        return output


from lighter_zoo import SegResNet
from monai.transforms import Compose, LoadImage, EnsureType, ScaleIntensityRange, Orientation, Resize
from monai.data import Dataset, DataLoader
import torch.nn as nn
import numpy as np
import torch


def load_model(model_path : str = "project-lighter/whole_body_segmentation"):
    # Load pre-trained model
    return SegResNet.from_pretrained(
        model_path
    )


def load_data(data): 
    train_transforms = Compose([ 
        EnsureType(),                        
        Orientation(axcodes="SPL"),       
        ScaleIntensityRange(
            a_min=-1024,    
            a_max=2048,    
            b_min=0,        
            b_max=1,        
            clip=True       
        ),
     ])

    train_ds = Dataset(data=data, transform=train_transforms)
    train_loader = DataLoader(train_ds, batch_size=1, shuffle=False)
    return train_loader

def encode(model, patches): 
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    adaptive_pool = nn.AdaptiveAvgPool3d((1, 1, 1))

    #expand patch to match encoder input requirements
    patch_array = np.expand_dims(patches, axis=(0, 1))
    
    train_loader = load_data(patch_array)

    model.eval() 
    with torch.no_grad():
        for input in train_loader:
            input = input.to(device)
            output = model.encoder(input) 
            #average pool and flatten the output to fit feature vector requirements
            output_flat = adaptive_pool(output[-1])
            out_flat = output_flat.flatten(start_dim =0)

    return out_flat.cpu().detach().numpy().tolist()



import logging
from typing import Any, Callable, Optional, Union

import torch
import torch.nn.functional as F
from torch import Tensor, einsum, nn
from transformers import PreTrainedModel
from transformers.generation.utils import GenerateOutput
from transformers.utils import add_start_docstrings, add_start_docstrings_to_model_forward

from unicorn_baseline.vision_language.prism.biogpt import BioGPT
from unicorn_baseline.vision_language.prism.configuring_prism import PrismConfig
from unicorn_baseline.vision_language.prism.perceiver import PerceiverResampler

logger = logging.getLogger(__file__)


class EmbedToLatents(nn.Module):
    def __init__(self, dim: int, dim_latents: int):
        super().__init__()

        self.to_latents = nn.Linear(dim, dim_latents, bias=False)

    def forward(self, x: Tensor) -> Tensor:
        latents = self.to_latents(x)
        return F.normalize(latents, dim=-1)


PRISM_DOCSTRING = """
    Prism is a vision-language model for histopathology.

    This model is a PyTorch [torch.nn.Module](https://pytorch.org/docs/stable/nn.html#torch.nn.Module) sub-class. Use
    it as a regular PyTorch Module and refer to the PyTorch documentation for all matter related to general usage and
    behavior.

    Parameters:
        config ([`~PrismConfig`]): Model configuration class with all the parameters of the model.
            Initializing with a config file does not load the weights associated with the model, only the
            configuration. Check out the [`~PreTrainedModel.from_pretrained`] method to load the model weights.
    """


PRISM_FORWARD_DOCSTRING = """
    Args:
        input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
            Indices of input sequence tokens in the vocabulary.
            Call `self.tokenize` method to obtain token indices.

        tile_embeddings (`torch.FloatTensor` of shape `(batch_size, tile_sequence_length, tile_embedding_dim)`):
            Tile embeddings for a whole slide image from Virchow V1 model.

            See: https://huggingface.co/paige-ai/Virchow

        tile_mask (`torch.LongTensor` of shape `(batch_size, tile_sequence_length)`):
            Mask to avoid performing attention on padding tiles. Mask values selected in `[0, 1]`:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.

            To do batch processing on multiple whole slide images, pad each tile embedding sequence to the max sequence
            length in the batch and concatenate into a tensor.

            Necessary since whole slide images can have different number of tiles and therefore `tile_sequence_length`
            will be different for each image.

            ```python
            image_0 = torch.randn(10, 2560)  # WSI with 10 tiles
            image_1 = torch.randn(15, 2560)  # WSI with 15 tiles

            max_len = max([len(image_0), len(image_1)])

            tile_embeddings = torch.zeros((2, max_len, 2560))
            tile_embeddings[0, :len(image_0)] = image_0
            tile_embeddings[1, :len(image_1)] = image_1

            tile_mask = torch.zeros((2, max_len))
            tile_mask[0, :len(image_0)] = torch.ones(len(image_0))
            tile_mask[1, :len(image_1)] = torch.ones(len(image_1))
            ```

        attention_mask (`torch.FloatTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Mask to avoid performing attention on padding token indices. Mask values selected in `[0, 1]`:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.

            If not provided, the mask will be computed using positions of pad token in `input_ids`.

        head_mask (`torch.FloatTensor` of shape `(num_heads,)` or `(num_layers, num_heads)`, *optional*):
            Mask to nullify selected heads of the self-attention modules. Mask values selected in `[0, 1]`:

            - 1 indicates the head is **not masked**,
            - 0 indicates the head is **masked**.

        past_key_values (`tuple(tuple(torch.FloatTensor))`, *optional*, returned when `use_cache=True` is passed or when `config.use_cache=True`):
            Tuple of `tuple(torch.FloatTensor)` of length `config.n_layers`, with each tuple having 2 tensors of shape
            `(batch_size, num_heads, sequence_length, embed_size_per_head)`) and 2 additional tensors of shape
            `(batch_size, num_heads, encoder_sequence_length, embed_size_per_head)`.

            Contains pre-computed hidden-states (key and values in the self-attention blocks and in the cross-attention
            blocks) that can be used (see `past_key_values` input) to speed up sequential decoding.

            If `past_key_values` are used, the user can optionally input only the last `decoder_input_ids` (those that
            don't have their past key value states given to this model) of shape `(batch_size, 1)` instead of all
            `decoder_input_ids` of shape `(batch_size, sequence_length)`. inputs_embeds (`torch.FloatTensor` of shape
            `(batch_size, sequence_length, hidden_size)`, *optional*): Optionally, instead of passing `input_ids` you
            can choose to directly pass an embedded representation. This is useful if you want more control over how to
            convert `input_ids` indices into associated vectors than the model's internal embedding lookup matrix.

        use_cache (`bool`, *optional*):
            If set to `True`, `past_key_values` key value states are returned and can be used to speed up decoding (see
            `past_key_values`).
    """


@add_start_docstrings(PRISM_DOCSTRING)
class Prism(PreTrainedModel):
    config_class = PrismConfig

    def __init__(self, config: PrismConfig):
        super().__init__(config)

        # Slide Encoder (Perceiver)

        self.image_resampler = PerceiverResampler(config.perceiver_config)

        # Text Decoder (BioGPT)

        self.text_decoder = BioGPT(
            config=config.biogpt_config,
            model_dir=config.model_dir,
            frozen_weights=config.biogpt_frozen_weights,
            frozen_embeddings=config.biogpt_frozen_embeddings,
            context_dim=config.biogpt_context_dim,
        )

        # Contrastive Head

        self.img_to_latents = EmbedToLatents(self.image_resampler.query_dim, config.dim_latents)
        self.text_to_latents = EmbedToLatents(self.text_decoder.text_dim, config.dim_latents)

        self.temperature = nn.Parameter(torch.log(torch.tensor(1.0) / 0.07))

        self.post_init()

    __call__: Callable[..., dict[str, Tensor]]

    @add_start_docstrings_to_model_forward(PRISM_FORWARD_DOCSTRING)
    def forward(
        self,
        input_ids: Tensor,
        tile_embeddings: Tensor,
        tile_mask: Optional[Tensor] = None,
        attention_mask: Optional[Tensor] = None,
        head_mask: Optional[Tensor] = None,
        past_key_values: Optional[tuple[tuple[Tensor]]] = None,
        use_cache: bool = False,
    ) -> dict[str, Tensor]:
        # perceiver

        resampler_out = self.image_resampler(
            tile_embeddings=tile_embeddings,
            tile_mask=tile_mask,
        )

        # biogpt

        decoder_out = self.text_decoder(
            input_ids=input_ids,
            key_value_states=resampler_out['image_latents'],
            attention_mask=attention_mask,
            head_mask=head_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
        )

        # embeddings to projections on a hypersphere

        text_proj = self.text_to_latents(decoder_out['text_embedding'])
        image_proj = self.img_to_latents(resampler_out['image_embedding'])

        # text-image similarity

        sim = einsum('i d, j d -> i j', text_proj, image_proj)
        sim = sim * self.temperature.exp()
        assert sim.shape[0] == sim.shape[1]

        return {
            'logits': decoder_out['logits'],
            'text_embedding': decoder_out['text_embedding'],
            'image_embedding': resampler_out['image_embedding'],
            'image_latents': resampler_out['image_latents'],
            'sim': sim,
        }

    def tokenize(self, text: list[str]) -> Tensor:
        return self.text_decoder.tokenize(text)

    def untokenize(self, token_ids: Tensor) -> list[str]:
        return self.text_decoder.untokenize(token_ids)

    @torch.no_grad()  # type: ignore
    def generate(
        self,
        inputs: Optional[Tensor] = None,
        key_value_states: Optional[Tensor] = None,
        max_length: int = 100,
        do_sample: bool = False,
        num_beams: int = 1,
        num_beam_groups: int = 1,
        **kwargs: Any,
    ) -> Union[GenerateOutput, torch.LongTensor]:
        """
        Args:
            inputs: input token indices used as prompt; leave as default (`None`) or
                use `"Diagnosis: "` string as prompt.
                Call `self.tokenize` to tokenize a batch of strings into indices.
                shape: (batch size, sequence length)

            key_value_states: image latents from Perceiver.
                Call `self.image_resampler(tile_embeddings, tile_mask)['image_latents']` to generate image latents.
                shape: (batch size, 512, 1280)

            max_length: the maximum length the generated tokens can have.

            do_sample: whether or not to use sampling; use greedy decoding otherwise.

            num_beams: number of beams for beam search. 1 means no beam search.

            num_beam_groups: number of groups to divide `num_beams` into in order to ensure
                diversity among different groups of beams.
                [this paper](https://arxiv.org/pdf/1610.02424.pdf) for more details.
        """
        if key_value_states is None:
            raise Exception('image latents (key_value_states) are required for generation.')

        batch_size = len(key_value_states)
        device = key_value_states.device

        if inputs is None:
            inputs = torch.tensor([[self.text_decoder.bos_token_id]] * batch_size, device=device)

        return self.text_decoder.model.generate(
            inputs=inputs,
            key_value_states=key_value_states,
            max_length=max_length,
            do_sample=do_sample,
            num_beams=num_beams,
            num_beam_groups=num_beam_groups,
            **kwargs,
        )

    @torch.no_grad()  # type: ignore
    def zero_shot(
        self,
        image_embeds: Tensor,
        *,
        neg_prompts: list[str],
        pos_prompts: list[str],
    ) -> Tensor:
        """
        Args:
            image_embeds: slide embedding per image.
                Call `self.image_resampler(tile_embeddings, tile_mask)['image_embedding']` to generate slide embeddings.
                shape: (image batch size, 1280)

            neg_prompts: list of prompts that are attributed to the negative class.
                If more than one is provided, their probability scores will be marginalised over.

            pos_prompts: list of prompts that are attributed to the positive class.
                If more than one is provided, their probability scores will be marginalised over.

        Returns:
            Probability values for the negative class and the positive class, per each image.
            shape: (image batch size, 2)
        """
        device = image_embeds.device

        # zero-shot prompts

        # (Bn + Bp, N)
        zero_shot_prompts = neg_prompts + pos_prompts
        zero_shot_token_ids = self.tokenize(zero_shot_prompts)[:, :-1]
        zero_shot_token_ids = zero_shot_token_ids.to(device)

        # (Bn + Bp, M)
        dummy_image_latents = torch.empty(
            (len(zero_shot_prompts), 1, self.text_decoder.context_dim), device=device
        )

        decoder_out = self.text_decoder(zero_shot_token_ids, dummy_image_latents)

        # zero-shot probabilities

        text_proj = self.text_to_latents(decoder_out['text_embedding'])
        image_proj = self.img_to_latents(image_embeds)

        # (Bi, Bn + Bp)
        sim = einsum('i d, j d -> i j', image_proj, text_proj)  # (image, text)
        sim = sim * self.temperature.exp()

        assert sim.shape[0] == len(image_embeds)
        assert sim.shape[1] == len(zero_shot_prompts)

        zero_shot_probs = torch.softmax(sim.to(torch.float), dim=-1)

        # (Bi, 2)
        neg_pos_probs = torch.cat(
            [
                zero_shot_probs[:, : len(neg_prompts)].sum(-1, keepdim=True),
                zero_shot_probs[:, len(neg_prompts) :].sum(-1, keepdim=True),
            ],
            dim=-1,
        )

        return neg_pos_probs

    def slide_representations(
        self,
        tile_embeddings: Tensor,
        tile_mask: Optional[Tensor] = None,
    ) -> dict[str, Tensor]:
        """
        Args:
            tile_embeddings: sequences of tile embeddings per whole slide image
                shape: (batch size, sequence length, feature dim)
            attn_mask: mask out padded area for each sequence in the batch;
                necessary for batch size > 1 since images can have different sequence lengths.
                shape: (batch size, sequence length)

        Returns:
            dict with image embedding and image latents per sequence
        """
        return self.image_resampler(tile_embeddings, tile_mask=tile_mask)



import sys
from contextlib import nullcontext
from pathlib import Path
from typing import Optional

import torch
import torch.nn as nn
import tqdm

from unicorn_baseline.vision.pathology.dataset import TileDataset, TileDatasetFromDisk


def extract_tile_features(
    *,
    model: nn.Module,
    dataloader: torch.utils.data.DataLoader,
    batch_size: int,
    device: torch.device,
    use_mixed_precision: bool,
):
    """Extracts tile-level features using the given model."""
    features = []
    autocast_context = nullcontext()
    if use_mixed_precision:
        autocast_context = torch.autocast(device_type="cuda", dtype=torch.float16)
    with torch.inference_mode():
        with autocast_context:
            for _, image in tqdm.tqdm(
                dataloader,
                desc="Extracting tile-level features",
                unit=" tile",
                unit_scale=batch_size,
                leave=True,
                file=sys.stdout,
            ):
                image = image.to(device, non_blocking=True)
                feature = model(image)
                features.append(feature.cpu())
    return torch.cat(features, dim=0)


def aggregate_slide_features(
    *,
    model: nn.Module,
    tile_features: torch.Tensor,
    dataset: torch.utils.data.Dataset,
    device: torch.device,
    use_mixed_precision: bool,
):
    """Aggregates tile-level features into a slide-level embedding."""
    tile_features = tile_features.to(device)
    coordinates = torch.tensor(dataset.coordinates, dtype=torch.int64, device=device)
    autocast_context = nullcontext()
    if use_mixed_precision:
        autocast_context = torch.autocast(device_type="cuda", dtype=torch.float16)
    with torch.inference_mode():
        with autocast_context:
            wsi_feature = model.forward_slide(
                tile_features,
                coordinates,
                tile_size_lv0=dataset.tile_size_lv0,
            )
    return wsi_feature.squeeze(0).cpu().tolist()


def extract_features(
    *,
    wsi_path: Path,
    model: nn.Module,
    coordinates_dir: Path,
    task_type: str,
    backend: str = "asap",
    batch_size: int = 1,
    num_workers: int = 4,
    use_mixed_precision: bool = False,
    load_tiles_from_disk: bool = False,
    tile_dir: Optional[str] = None,
    tile_format: Optional[str] = None,
):
    """Main function to extract features for all cases."""

    model.eval().to(model.device)
    transforms = model.get_transforms()

    if load_tiles_from_disk:
        dataset = TileDatasetFromDisk(
            wsi_path,
            tile_dir=tile_dir,
            tile_format=tile_format,
            transforms=transforms,
        )
    else:
        dataset = TileDataset(
            wsi_path, coordinates_dir=coordinates_dir, backend=backend, transforms=transforms
        )

    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=True
    )

    tile_features = extract_tile_features(
        model=model,
        dataloader=dataloader,
        batch_size=batch_size,
        device=model.device,
        use_mixed_precision=use_mixed_precision,
    )

    if task_type in ["classification", "regression"]:
        # aggregate to a single slide-level feature
        slide_feature = aggregate_slide_features(
            model=model, tile_features=tile_features, dataset=dataset, device=model.device, use_mixed_precision=use_mixed_precision
        )
        return slide_feature
    else:
        # return tile-level features for detection/segmentation tasks
        return tile_features



#  Copyright 2025 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from glob import glob
from pathlib import Path
from typing import Any

import numpy as np


def load_task_description(
    input_path: Path = Path("/input/unicorn-task-description.json"),
) -> dict[str, str]:
    """
    Read information from unicorn-task-description.json
    """
    with open(input_path, "r") as f:
        task_description = json.load(f)
    return task_description


def load_inputs(input_path: Path = Path("/input/inputs.json")) -> list[dict[str, Any]]:
    """
    Read information from inputs.json
    """
    input_information_path = Path(input_path)
    with input_information_path.open("r") as f:
        input_information = json.load(f)

    for item in input_information:
        relative_path = item["interface"]["relative_path"]
        item["input_location"] = Path(f"/input/{relative_path}")

    return input_information


def sanitize_json_content(obj):
    if isinstance(obj, dict):
        return {k: sanitize_json_content(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple, np.ndarray)):
        return [sanitize_json_content(v) for v in obj]
    elif isinstance(obj, (str, int, bool, float)):
        return obj
    elif isinstance(obj, (np.float16, np.float32, np.float64)):
        return float(obj)
    elif isinstance(obj, (np.uint8, np.uint16, np.uint32, np.uint64, np.int8, np.int16, np.int32, np.int64)):
        return int(obj)
    else:
        return obj.__repr__()


def write_json_file(*, location, content):
    # Writes a json file with the sanitized content
    content = sanitize_json_content(content)
    with open(location, "w") as f:
        f.write(json.dumps(content, indent=4))


def resolve_image_path(*, location: str | Path) -> Path:
    input_files = (
        glob(str(location / "*.tif"))
        + glob(str(location / "*.tiff"))
        + glob(str(location / "*.mha"))
    )
    if len(input_files) != 1:
        raise ValueError(f"Expected one image file, got {len(input_files)}")

    input_file = Path(input_files[0])
    return input_file


if __name__ == "__main__":
    print(load_inputs("baseline_template/test/input/inputs.json"))
    print(
        load_task_description(
            "baseline_template/test/input/unicorn-task-description.json"
        )
    )



#  Copyright 2025 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from pathlib import Path
from typing import Any

from unicorn_baseline.vision.pathology.main import run_pathology_vision_task
from unicorn_baseline.vision.radiology.main import run_radiology_vision_task


def run_vision(
    task_description: dict[str, str],
    input_information: list[dict[str, Any]],
    model_dir: Path,
) -> int:
    """
    Process input data
    """
    # retrieve task details
    domain = task_description["domain"]
    task_type = task_description["task_type"]

    if domain == "pathology":
        run_pathology_vision_task(
            task_type=task_type,
            input_information=input_information,
            model_dir=model_dir,
        )
    elif domain == "radiology":
        run_radiology_vision_task(
            task_type=task_type,
            input_information=input_information,
            model_dir=model_dir,
        )
    else:
        raise ValueError(f"Domain '{domain}' not supported yet for vision tasks.")

    return 0



'''
    Adapted from https://github.com/bytedance/ibot/blob/da316d82636a7a7356835ef224b13d5f3ace0489/models/vision_transformer.py and timm (https://github.com/huggingface/pytorch-image-models) v0.9.2
'''

import logging
import math
from functools import partial
from typing import Callable, List, Optional, Sequence, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint
from torch.jit import Final
from torch import einsum

import torchvision.transforms as T

from enum import Enum
from einops import rearrange, repeat
from einops_exts import rearrange_many

# Imports called in timm's vision_transformer.py
from timm.layers.helpers import to_2tuple
from timm.layers import Mlp, DropPath, trunc_normal_, PatchDropout, use_fused_attn
from timm.models._manipulate import named_apply, checkpoint_seq

# Imports of other functions called within timm's vision_transformer.py (that are not defined below)
from timm.models.vision_transformer import init_weights_vit_timm, get_init_weights_vit, _load_weights

from unicorn_baseline.vision.pathology.model_utils import update_state_dict

class Format(str, Enum):
    NCHW = 'NCHW'
    NHWC = 'NHWC'
    NCL = 'NCL'
    NLC = 'NLC'

def nchw_to(x: torch.Tensor, fmt: Format):
    if fmt == Format.NHWC:
        x = x.permute(0, 2, 3, 1)
    elif fmt == Format.NLC:
        x = x.flatten(2).transpose(1, 2)
    elif fmt == Format.NCL:
        x = x.flatten(2)
    return x

class PatchEmbed(nn.Module):
    """ 2D Image to Patch Embedding
    """
    output_fmt: Format

    def __init__(
            self,
            img_size: int = 224,
            patch_size: int = 16,
            in_chans: int = 3,
            embed_dim: int = 768,
            norm_layer: Optional[Callable] = None,
            flatten: bool = True,
            output_fmt: Optional[str] = None,
            bias: bool = True,
            masked_im_modeling: bool = False
    ):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.grid_size = (img_size[0] // patch_size[0], img_size[1] // patch_size[1])
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        if output_fmt is not None:
            self.flatten = False
            self.output_fmt = Format(output_fmt)
        else:
            # flatten spatial dim and transpose to channels last, kept for bwd compat
            self.flatten = flatten
            self.output_fmt = Format.NCHW

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size, bias=bias)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

        ### Mask Image Modeling
        self.masked_im_modeling = masked_im_modeling
        if self.masked_im_modeling:
            self.masked_embed = nn.Parameter(torch.zeros(1, embed_dim))

    def forward(self, x, mask=None):
        B, C, H, W = x.shape
        # iBOT Modification: Commented out asserts
        #_assert(H == self.img_size[0], f"Input image height ({H}) doesn't match model ({self.img_size[0]}).")
        #_assert(W == self.img_size[1], f"Input image width ({W}) doesn't match model ({self.img_size[1]}).")
        x = self.proj(x)

        if mask is not None:
            x = self.mask_model(x, mask)

        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # NCHW -> NLC
        elif self.output_fmt != Format.NCHW:
            x = nchw_to(x, self.output_fmt)
        x = self.norm(x)
        return x
    
    def mask_model(self, x, mask):
        x.permute(0, 2, 3, 1)[mask, :] = self.masked_embed.to(x.dtype)
        return x

### No Major Change from 0.8.13dev0 -> 9.2.0
# fast attn -> fused attn naming
class Attention(nn.Module):
    fused_attn: Final[bool]

    def __init__(
            self,
            dim,
            num_heads=8,
            qkv_bias=False,
            qk_norm=False,
            attn_drop=0.,
            proj_drop=0.,
            norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim ** -0.5
        self.fused_attn = use_fused_attn()
        self.fast_attn = self.fused_attn # legacy support

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.q_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.k_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, return_attention=False):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)
        q, k = self.q_norm(q), self.k_norm(k)

        if self.fused_attn and (return_attention == False):
            x = F.scaled_dot_product_attention(
                q, k, v,
                dropout_p=self.attn_drop.p,
            )
        else:
            q = q * self.scale
            attn = q @ k.transpose(-2, -1)
            attn = attn.softmax(dim=-1)
            attn = self.attn_drop(attn)
            x = attn @ v

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        if return_attention:
            return x, attn
        return x

### No Change from 0.8.13dev0 -> 9.2.0
class LayerScale(nn.Module):
    def __init__(self, dim, init_values=1e-5, inplace=False):
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x):
        return x.mul_(self.gamma) if self.inplace else x * self.gamma

### No Major Change from 0.8.13dev0 -> 9.2.0
# drop renamed as proj_drop
# Mlp layer added
class Block(nn.Module):
    def __init__(
            self,
            dim,
            num_heads,
            mlp_ratio=4.,
            qkv_bias=False,
            qk_norm=False,
            proj_drop=0., # proj -> proj_drop renamed
            attn_drop=0.,
            init_values=None,
            drop_path=0.,
            act_layer=nn.GELU,
            norm_layer=nn.LayerNorm,
            mlp_layer=Mlp,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_norm=qk_norm,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
            norm_layer=norm_layer,
        )
        self.ls1 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        self.drop_path1 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.norm2 = norm_layer(dim)
        self.mlp = mlp_layer( ### Mlp -> mlp_layer
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            act_layer=act_layer,
            drop=proj_drop,
        )
        self.ls2 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        self.drop_path2 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x):
        x = x + self.drop_path1(self.ls1(self.attn(self.norm1(x))))
        x = x + self.drop_path2(self.ls2(self.mlp(self.norm2(x))))
        return x
    
    def forward_with_attention(self, x):
        x_input = x
        x_postattn, attn = self.attn(self.norm1(x_input), return_attention=True)
        x_postls1 = x_input + self.drop_path1(self.ls1(x_postattn))
        x_postmlp = self.mlp(self.norm2(x_postls1))
        x_postls2 = x_postls1 + self.drop_path2(self.ls2(x_postmlp))
        return x_postls2, attn


class VisionTransformer(nn.Module):
    """ Vision Transformer
    A PyTorch impl of : `An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale`
        - https://arxiv.org/abs/2010.11929
    Adapted entirely from: https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/vision_transformer.py
    Last updated version: 0.8.17dev0
    """

    def __init__(
            self,
             img_size: Union[int, Tuple[int, int]] = 224,
            patch_size: Union[int, Tuple[int, int]] = 16,
            in_chans: int = 3,
            num_classes: int = 0,
            global_pool: str = 'token',
            embed_dim: int = 768,
            depth: int = 12,
            num_heads: int = 12,
            mlp_ratio: float = 4.,
            qkv_bias: bool = True,
            qk_norm: bool = False,
            init_values: Optional[float] = None,
            class_token: bool = True,
            no_embed_class: bool = False,
            pre_norm: bool = False,
            fc_norm: Optional[bool] = None,
            drop_rate: float = 0.,
            pos_drop_rate: float = 0.,   # new
            patch_drop_rate: float = 0., # new
            proj_drop_rate: float = 0.,  # renamed
            attn_drop_rate: float = 0.,  # same
            drop_path_rate: float = 0.,  # same
            weight_init: str = '',
            embed_layer: Callable = PatchEmbed,
            norm_layer: Optional[Callable] = None,
            act_layer: Optional[Callable] = None,
            block_fn: Callable = Block,     
            mlp_layer: Callable = Mlp,  # New
            return_all_tokens=False,    # iBOT
            masked_im_modeling=False    # iBOT
    ):
        """
        Args:
            img_size: Input image size.
            patch_size: Patch size.
            in_chans: Number of image input channels.
            num_classes: Mumber of classes for classification head.
            global_pool: Type of global pooling for final sequence (default: 'token').
            embed_dim: Transformer embedding dimension.
            depth: Depth of transformer.
            num_heads: Number of attention heads.
            mlp_ratio: Ratio of mlp hidden dim to embedding dim.
            qkv_bias: Enable bias for qkv projections if True.
            init_values: Layer-scale init values (layer-scale enabled if not None).
            class_token: Use class token.
            fc_norm: Pre head norm after pool (instead of before), if None, enabled when global_pool == 'avg'.
            drop_rate: Head dropout rate.
            pos_drop_rate: Position embedding dropout rate.
            attn_drop_rate: Attention dropout rate.
            drop_path_rate: Stochastic depth rate.
            weight_init: Weight initialization scheme.
            embed_layer: Patch embedding layer.
            norm_layer: Normalization layer.
            act_layer: MLP activation layer.
            block_fn: Transformer block layer.
        """
        super().__init__()
        assert global_pool in ('', 'avg', 'token')
        assert class_token or global_pool != 'token'
        use_fc_norm = global_pool == 'avg' if fc_norm is None else fc_norm
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        act_layer = act_layer or nn.GELU
        self.return_all_tokens = return_all_tokens     # from ibot
        self.masked_im_modeling = masked_im_modeling   # from ibot

        self.num_classes = num_classes
        self.global_pool = global_pool
        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models
        self.num_prefix_tokens = 1 if class_token else 0
        self.no_embed_class = no_embed_class
        self.grad_checkpointing = False

        self.patch_embed = embed_layer(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
            bias=not pre_norm,  # disable bias if pre-norm is used (e.g. CLIP)
            masked_im_modeling=masked_im_modeling,
        )
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim)) if class_token else None
        embed_len = num_patches if no_embed_class else num_patches + self.num_prefix_tokens
        self.pos_embed = nn.Parameter(torch.randn(1, embed_len, embed_dim) * .02)
        self.pos_drop = nn.Dropout(p=pos_drop_rate)
        # new
        if patch_drop_rate > 0:
            self.patch_drop = PatchDropout(
                patch_drop_rate,
                num_prefix_tokens=self.num_prefix_tokens,
            )
        else:
            self.patch_drop = nn.Identity()
        self.norm_pre = norm_layer(embed_dim) if pre_norm else nn.Identity()

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.Sequential(*[
            block_fn(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_norm=qk_norm,
                init_values=init_values,
                proj_drop=proj_drop_rate, # renamed
                attn_drop=attn_drop_rate,
                drop_path=dpr[i],
                norm_layer=norm_layer,
                act_layer=act_layer,
                mlp_layer=mlp_layer,  # new
            )
            for i in range(depth)])
        self.norm = norm_layer(embed_dim) if not use_fc_norm else nn.Identity()

        # Classifier Head
        self.fc_norm = norm_layer(embed_dim) if use_fc_norm else nn.Identity()
        self.head_drop = nn.Dropout(drop_rate) # new
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

        if weight_init != 'skip':
            self.init_weights(weight_init)

    def init_weights(self, mode=''):
        assert mode in ('jax', 'jax_nlhb', 'moco', '')
        head_bias = -math.log(self.num_classes) if 'nlhb' in mode else 0.
        trunc_normal_(self.pos_embed, std=.02)
        if self.cls_token is not None:
            nn.init.normal_(self.cls_token, std=1e-6)
        named_apply(get_init_weights_vit(mode, head_bias), self)

    def _init_weights(self, m):
        # this fn left here for compat with downstream users
        init_weights_vit_timm(m)

    @torch.jit.ignore()
    def load_pretrained(self, checkpoint_path, prefix=''):
        _load_weights(self, checkpoint_path, prefix)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed', 'cls_token', 'dist_token'}

    @torch.jit.ignore
    def group_matcher(self, coarse=False):
        return dict(
            stem=r'^cls_token|pos_embed|patch_embed',  # stem and embed
            blocks=[(r'^blocks\.(\d+)', None), (r'^norm', (99999,))]
        )

    @torch.jit.ignore
    def set_grad_checkpointing(self, enable=True):
        self.grad_checkpointing = enable

    @torch.jit.ignore
    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes: int, global_pool=None):
        self.num_classes = num_classes
        if global_pool is not None:
            assert global_pool in ('', 'avg', 'token')
            self.global_pool = global_pool
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

    def _pos_embed(self, x, w, h):
        if self.no_embed_class:
            # deit-3, updated JAX (big vision)
            # position embedding does not overlap with class token, add then concat
            x = x + self.interpolate_pos_encoding(x, w, h)
            if self.cls_token is not None:
                x = torch.cat((self.cls_token.expand(x.shape[0], -1, -1), x), dim=1)
        else:
            # original timm, JAX, and deit vit impl
            # pos_embed has entry for class token, concat then add
            if self.cls_token is not None:
                x = torch.cat((self.cls_token.expand(x.shape[0], -1, -1), x), dim=1)
            x = x + self.interpolate_pos_encoding(x, w, h)
        return self.pos_drop(x)

    ### all new
    def _intermediate_layers(
            self,
            x: torch.Tensor,
            n: Union[int, Sequence] = 1,
    ):
        outputs, num_blocks = [], len(self.blocks)
        take_indices = set(range(num_blocks - n, num_blocks) if isinstance(n, int) else n)

        # forward pass
        x = self.patch_embed(x)
        x = self._pos_embed(x)
        x = self.patch_drop(x)
        x = self.norm_pre(x)
        for i, blk in enumerate(self.blocks):
            x = blk(x)
            if i in take_indices:
                outputs.append(x)

        return outputs

    def get_intermediate_layers(
            self,
            x: torch.Tensor,
            n: Union[int, Sequence] = 1,
            reshape: bool = False,
            return_class_token: bool = False,
            norm: bool = False,
    ) -> Tuple[Union[torch.Tensor, Tuple[torch.Tensor]]]:
        """ Intermediate layer accessor (NOTE: This is a WIP experiment).
        Inspired by DINO / DINOv2 interface
        """
        # take last n blocks if n is an int, if in is a sequence, select by matching indices
        outputs = self._intermediate_layers(x, n)
        if norm:
            outputs = [self.norm(out) for out in outputs]
        class_tokens = [out[:, 0:self.num_prefix_tokens] for out in outputs]
        outputs = [out[:, self.num_prefix_tokens:] for out in outputs]

        if reshape:
            grid_size = self.patch_embed.grid_size
            outputs = [
                out.reshape(x.shape[0], grid_size[0], grid_size[1], -1).permute(0, 3, 1, 2).contiguous()
                for out in outputs
            ]

        if return_class_token:
            return tuple(zip(outputs, class_tokens))
        return tuple(outputs)
    
    def forward_features(self, x, mask=None):
        B, nc, w, h = x.shape
        
        ### Specific to iBOT
        if self.masked_im_modeling:
            assert mask is not None
            x = self.patch_embed(x, mask) if self.masked_im_modeling else self.patch_embed(x)
        else:
            x = self.patch_embed(x)
        
        x = self._pos_embed(x, w, h)
        x = self.patch_drop(x)
        x = self.norm_pre(x)
        if self.grad_checkpointing and not torch.jit.is_scripting():
            x = checkpoint_seq(self.blocks, x)
        else:
            x = self.blocks(x)
        x = self.norm(x)
        return x

    def forward_head(self, x, pre_logits: bool = False):
        if self.global_pool:
            x = x[:, self.num_prefix_tokens:].mean(dim=1) if self.global_pool == 'avg' else x[:, 0]
        x = self.fc_norm(x)
        x = self.head_drop(x) # new
        return x if pre_logits else self.head(x)
    
    def get_attention(self, x, block_num: int=-1):
        B, nc, w, h = x.shape
        x = self.patch_embed(x)
        x = self._pos_embed(x, w, h)
        x = self.patch_drop(x)
        x = self.norm_pre(x)

        if block_num < 0:
            block_num = len(self.blocks) + block_num
        
        if self.grad_checkpointing and not torch.jit.is_scripting():
            raise NotImplementedError
        else:
            for i, blk in enumerate(self.blocks):
                if i < block_num:
                    x = blk(x)
                else:
                    x, attn = blk.forward_with_attention(x)
                    return attn

    def forward(self, x, return_all_tokens=None, mask=None):
        if self.masked_im_modeling:
            assert mask is not None
            x = self.forward_features(x, mask)
        else:
            x = self.forward_features(x)
        
        ### iBOT
        return_all_tokens = self.return_all_tokens if \
            return_all_tokens is None else return_all_tokens
        if return_all_tokens:
            return x
        
        x = self.forward_head(x)
        return x
    
    
    def interpolate_pos_encoding(self, x, w, h):
        npatch = x.shape[1] - 1
        N = self.pos_embed.shape[1] - 1
        if npatch == N and w == h:
            return self.pos_embed
        class_pos_embed = self.pos_embed[:, 0]
        patch_pos_embed = self.pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_embed.patch_size[0]
        h0 = h // self.patch_embed.patch_size[0]
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        w0, h0 = w0 + 0.1, h0 + 0.1
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
            scale_factor=(w0 / math.sqrt(N), h0 / math.sqrt(N)),
            mode='bicubic',
        )
        assert int(w0) == patch_pos_embed.shape[-2] and int(h0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1)

# see deprecation notice below:
def resize_pos_embed(model, pos_embed_w):
    resized = False
    if pos_embed_w.shape != model.pos_embed.shape:
        # see https://github.com/rwightman/pytorch-image-models/blob/624266148d8fa5ddb22a6f5e523a53aaf0e8a9eb/timm/models/vision_transformer.py#L509
        interpolation = 'bilinear'
        antialias = False
        try:
            from timm.layers import resample_abs_pos_embed
        except:
            print(f'{__file__}: import timm utility functions failed with version {timm.__version__}!')
        num_prefix_tokens = 0 if getattr(model, 'no_embed_class', False) else getattr(model, 'num_prefix_tokens', 1)
        pos_embed_w = resample_abs_pos_embed(  # resize pos embedding when different size from pretrained weights
                    pos_embed_w,
                    new_size=model.patch_embed.grid_size,
                    num_prefix_tokens=num_prefix_tokens,
                    interpolation=interpolation,
                    antialias=antialias,
                    verbose=True,
                )
        
        resized = True
    if not resized:
        logging.info('pos embedding not resized.')
    return pos_embed_w

class AttentionalPooler(nn.Module):
    def __init__(
            self,
            d_model: int,
            context_dim: int,
            n_head: int = 8,
            n_queries: int = 256,
            norm_layer: Callable = nn.LayerNorm
    ):
        super().__init__()
        self.query = nn.Parameter(torch.randn(n_queries, d_model))
        dim_head = d_model // n_head

        self.scale = dim_head ** -0.5
        self.heads = n_head
        inner_dim = dim_head * n_head

        self.ln_k = norm_layer(context_dim)
        self.ln_q = norm_layer(d_model)

        self.to_q = nn.Linear(d_model, inner_dim, bias=False)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False)
        self.to_out = nn.Linear(inner_dim, d_model, bias=False)

    def forward(self, x: torch.Tensor, attn_mask: Optional[torch.Tensor] = None):
        if x.ndim == 3:
            x = rearrange(x, 'b n d -> b 1 n d')

        q = repeat(self.query, 'n d -> b m n d', b=x.shape[0], m=x.shape[1])

        x = self.ln_k(x)
        q = self.ln_q(q)
        b, m, h = *x.shape[:2], self.heads

        q = self.to_q(q)

        kv_input = x
        k, v = self.to_kv(kv_input).chunk(2, dim=-1)

        q, k, v = rearrange_many((q, k, v), 'b t n (h d) -> b h t n d', h=h)

        q = q * self.scale

        # attention
        sim = einsum('... i d, ... j d  -> ... i j', q, k)

        sim = sim - sim.amax(dim=-1, keepdim=True).detach()
        attn = sim.softmax(dim=-1)

        out = einsum('... i j, ... j d -> ... i d', attn, v)
        out = rearrange(out, 'b h t n d -> b t n (h d)', h=h)
        return self.to_out(out).squeeze(dim=1)

class EncoderWithAttentionalPooler(nn.Module):
    def __init__(
        self,
        encoder,
        attn_pooler_contrast,
        embed_dim,
        norm_layer: Callable = nn.LayerNorm,
        global_average_pool: bool = False
    ):
        super().__init__()
        self.trunk = encoder
        self.attn_pool_contrast = attn_pooler_contrast
        self.ln_contrast = norm_layer(embed_dim)
        self.global_average_pool = global_average_pool
    
    def _global_pool(self, x):
        if self.global_average_pool:
            return x.mean(dim=1), x
        else:
            return x[:, 0], x[:, 1:]
    
    def forward(self, x):
        x = self.trunk(x, return_all_tokens=True)
        if self.global_average_pool:
            pooled, _ = self._global_pool(x)
        else:
            pooled = self.attn_pool_contrast(x)[:, 0]
            pooled = self.ln_contrast(pooled)
        return pooled

def build_conch(conch_cfg, model_dir):
    model = VisionTransformer(
        patch_size=conch_cfg.patch_size, 
        embed_dim=conch_cfg.context_dim, 
        depth=conch_cfg.depth, 
        num_heads=conch_cfg.num_heads, 
        mlp_ratio=conch_cfg.mlp_ratio,
        qkv_bias=conch_cfg.qkv_bias, 
        init_values=conch_cfg.init_values
    )
    attn_pooler_contrast = AttentionalPooler(d_model=conch_cfg.embed_dim, 
                                            context_dim=conch_cfg.context_dim, 
                                            n_queries=conch_cfg.pooler_n_queries_contrast)
    model = EncoderWithAttentionalPooler(encoder=model, 
                                         attn_pooler_contrast=attn_pooler_contrast, 
                                         embed_dim=conch_cfg.embed_dim)

    checkpoint_path = model_dir / "conch-tile-encoder.pth"
    print(f"Loading tile-level encoder from {checkpoint_path} ...")
    state_dict = torch.load(checkpoint_path, map_location="cpu")
    updated_sd, msg = update_state_dict(model_dict=model.state_dict(), state_dict=state_dict)
    model.load_state_dict(updated_sd, strict=True)
    print(msg)

    img_size = 448
    eval_transform = T.Compose([
            T.Resize(img_size, interpolation=T.InterpolationMode.BILINEAR),
            T.CenterCrop(img_size),
            T.ToTensor(),
            T.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225))
    ])
    return model, eval_transform


import setuptools

if __name__ == "__main__":
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    setuptools.setup(
        author_email="Clement.Grisi@radboudumc.nl",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/DIAGNijmegen/unicorn_baseline",
        project_urls={
            "Bug Tracker": "https://github.com/DIAGNijmegen/unicorn_baseline/issues"
        },
        package_dir={"": "src"},  # our packages live under src, but src is not a package itself
        packages=setuptools.find_packages("src", exclude=["tests"]),
        exclude_package_data={"": ["tests"]},
        entry_points={
            "console_scripts": [
                "unicorn_baseline = unicorn_baseline.inference:run",
            ],
        },
    )



#  Copyright 2025 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from typing import Any


def run_radiology_vision_task(
    *,
    task_type: str,
    input_information: dict[str, Any],
    model_dir: Path,
):
    raise NotImplementedError(
        "Radiology vision task is not implemented yet."
    )
